@@ -46,7 +46,7 @@ from hypothesis.internal.conjecture.datatree import DataTree, TreeRecordingObser
 from hypothesis.internal.conjecture.junkdrawer import pop_random, uniform
 from hypothesis.internal.conjecture.shrinker import Shrinker, sort_key
 from hypothesis.internal.healthcheck import fail_health_check
-from hypothesis.reporting import debug_report
+from hypothesis.reporting import base_report
 
 # Tell pytest to omit the body of this module from tracebacks
 # https://docs.pytest.org/en/latest/example/simple.html#writing-well-integrated-assertion-helpers
@@ -56,6 +56,7 @@ __tracebackhide__ = True
 MAX_SHRINKS = 500
 CACHE_SIZE = 10000
 MUTATION_POOL_SIZE = 100
+MIN_TEST_CALLS = 10
 
 
 @attr.s
@@ -98,6 +99,9 @@ class ConjectureRunner(object):
         self.target_selector = TargetSelector(self.random)
 
         self.interesting_examples = {}
+        # We use call_count because there may be few possible valid_examples.
+        self.first_bug_found_at = None
+        self.last_bug_found_at = None
 
         self.shrunk_examples = set()
 
@@ -182,6 +186,9 @@ class ConjectureRunner(object):
                 existing = self.interesting_examples[key]
             except KeyError:
                 changed = True
+                self.last_bug_found_at = self.call_count
+                if self.first_bug_found_at is None:
+                    self.first_bug_found_at = self.call_count
             else:
                 if sort_key(data.buffer) < sort_key(existing.buffer):
                     self.shrinks += 1
@@ -199,6 +206,10 @@ class ConjectureRunner(object):
                 self.exit_with(ExitReason.max_shrinks)
 
         if not self.interesting_examples:
+            # Note that this logic is reproduced to end the generation phase when
+            # we have interesting examples.  Update that too if you change this!
+            # (The doubled implementation is because here we exit the engine entirely,
+            #  while in the other case below we just want to move on to shrinking.)
             if self.valid_examples >= self.settings.max_examples:
                 self.exit_with(ExitReason.max_examples)
             if self.call_count >= max(
@@ -339,8 +350,8 @@ class ConjectureRunner(object):
             self.event_call_counts[event] += 1
 
     def debug(self, message):
-        with local_settings(self.settings):
-            debug_report(message)
+        if self.settings.verbosity >= Verbosity.debug:
+            base_report(message)
 
     @property
     def report_debug_info(self):
@@ -588,6 +599,11 @@ class ConjectureRunner(object):
     def generate_new_examples(self):
         if Phase.generate not in self.settings.phases:
             return
+        if self.interesting_examples:
+            # The example database has failing examples from a previous run,
+            # so we'd rather report that they're still failing ASAP than take
+            # the time to look for additional failures.
+            return
 
         zero_data = self.cached_test_function(hbytes(self.settings.buffer_size))
         if zero_data.status > Status.OVERRUN:
@@ -635,9 +651,42 @@ class ConjectureRunner(object):
 
         self.health_check_state = HealthCheckState()
 
+        def should_generate_more():
+            # If we haven't found a bug, keep looking.  We check this before
+            # doing anything else as it's by far the most common case.
+            if not self.interesting_examples:
+                return True
+            # If we've found a bug and won't report more than one, stop looking.
+            elif not self.settings.report_multiple_bugs:
+                return False
+            assert self.first_bug_found_at <= self.last_bug_found_at <= self.call_count
+            # End the generation phase where we would have ended it if no bugs had
+            # been found.  This reproduces the exit logic in `self.test_function`,
+            # but with the important distinction that this clause will move on to
+            # the shrinking phase having found one or more bugs, while the other
+            # will exit having found zero bugs.
+            if (
+                self.valid_examples >= self.settings.max_examples
+                or self.call_count >= max(self.settings.max_examples * 10, 1000)
+            ):  # pragma: no cover
+                return False
+            # Otherwise, keep searching for between ten and 'a heuristic' calls.
+            # We cap 'calls after first bug' so errors are reported reasonably
+            # soon even for tests that are allowed to run for a very long time,
+            # or sooner if the latest half of our test effort has been fruitless.
+            return self.call_count < MIN_TEST_CALLS or self.call_count < min(
+                self.first_bug_found_at + 1000, self.last_bug_found_at * 2
+            )
+
         count = 0
-        while not self.interesting_examples and (
-            count < 10 or self.health_check_state is not None
+        while should_generate_more() and (
+            count < 10
+            or self.health_check_state is not None
+            # If we have not found a valid prefix yet, the target selector will
+            # be empty and the mutation stage will fail with a very rare internal
+            # error.  We therefore continue this initial random generation step
+            # until we have found at least one prefix to mutate.
+            or len(self.target_selector) == 0
         ):
             prefix = self.generate_novel_prefix()
 
@@ -661,7 +710,7 @@ class ConjectureRunner(object):
 
         zero_bound_queue = []
 
-        while not self.interesting_examples:
+        while should_generate_more():
             if zero_bound_queue:
                 # Whenever we generated an example and it hits a bound
                 # which forces zero blocks into it, this creates a weird
@@ -763,6 +812,12 @@ class ConjectureRunner(object):
             )
             self.debug("Shrinking %r" % (target,))
 
+            if not self.settings.report_multiple_bugs:
+                # If multi-bug reporting is disabled, we shrink our currently-minimal
+                # failure, allowing 'slips' to any bug with a smaller minimal example.
+                self.shrink(example, lambda d: d.status == Status.INTERESTING)
+                return
+
             def predicate(d):
                 if d.status < Status.INTERESTING:
                     return False
